using System;
using System.Collections;
using System.IO;
using System.Text;
using System.Xml;
using Gregor.Core;
using Gregor.NetConsole;

namespace Gregor.NetSheet {

[Module()]
public class CellUtil {

    public static bool IsNumber(object o){
        double d = 0.0;
        return TryGetNumber(o, ref d);
    }

    public static bool TryGetNumber(object o, ref double d){
        if(o is int){
            d = (int)o; // unbox int, convert to double
            return true;
        }else if(o is long){
            d = (long)o; // unbox long, convert to double
            return true;
        }else if(o is double){
            d = (double)o; // unbox double
            return true;
        }else{
            return false;
        }
    }

    public static bool AreValidCoordinates(string sCoords){
        try{
            int iRow = 0, iCol = 0;
            ParseCoordinates(sCoords, ref iRow, ref iCol);
            return true;
        }catch(Exception){
            return false;
        }
    }

    public static void ParseCoordinates(string sCoords, ref int iRow, ref int iCol){
        Check.String(sCoords);
        bool fA1 = char.IsLetter(sCoords[0]);
        int iDot = sCoords.IndexOfAny(new char[]{'.', ',', ';', ':'});
        if(fA1){
            if(iDot >= 0){
                throw new ArgumentException("Coordinates in A1 format must not be separated by punctuation.");
            }
            CellUtil.ParseA1Coordinates(sCoords, ref iRow, ref iCol);
        }else{
            if(iDot < 0){
                throw new ArgumentException("Coordinates in 0.0 format must be separated by a dot, comma, colon, or semicolon.");
            }
            if(iDot == 0 || iDot == sCoords.Length - 1){
                throw new ArgumentException("Must specify two coordinates.");
            }
            CellUtil.Parse00Coordinates(sCoords, iDot, ref iRow, ref iCol);
        }
        if(iRow < 0 || iCol < 0){
            throw new ArgumentException("Coordinates cannot be negative.");
        }
    }

    private static void Parse00Coordinates(string sCoords, int iDot, ref int iRow, ref int iCol){
        char ch = sCoords[iDot];
        string[] aStr = sCoords.Split(ch);
        iRow = int.Parse(aStr[0]);
        iCol = int.Parse(aStr[1]);
    }

    private static void ParseA1Coordinates(string sCoords, ref int iRow, ref int iCol){
        StringBuilder sbLetters = new StringBuilder();
        int i = 0;
        while(i < sCoords.Length && char.IsLetter(sCoords[i])){
            sbLetters.Append(char.ToUpper(sCoords[i]));
            i++;
        }
        if(sbLetters.Length == 0){
            throw new ArgumentException("No column specified.");
        }
        if(sbLetters.Length > 2){
            throw new ArgumentException("Coordinates in A1 format are limited to 676 (ZZ) columns.");
        }
        if(sbLetters.Length < sCoords.Length){
            for(i = sbLetters.Length; i < sCoords.Length; i++){
                if(false == char.IsDigit(sCoords[i])){
                    throw new ArgumentException("Coordinates in A1 format must use digits for row.");
                }
            }
        }else{
            throw new ArgumentException("No row specified.");
        }
        // so
        int val = 0;
        int fac = 1;
        int ascA = (int) 'A';
        for(i = sbLetters.Length - 1; i >= 0; i--){
            int v = (int)sbLetters[i] - ascA + 1;
            val += v * fac;
            fac *= 26;
        }
        iCol = val - 1;
        iRow = int.Parse(sCoords.Substring(sbLetters.Length)) - 1;
    }

    public static string TranslateCoordinates(int iRow, int iCol){
        return TranslateRow(iRow) + TranslateColumn(iCol);
    }

    public static string TranslateRow(int iRow){
        return (iRow + 1).ToString();
    }

    public static string TranslateColumn(int iCol){
        if(iCol >= 676){
            throw new ArgumentException("Coordinates in A1 format are limited to 676 (ZZ) columns.");
        }else if(iCol < 26){
            char ch = Translate0To25(iCol);
            return ch.ToString();
        }else{
            int val = (iCol + 1);
            int i1 = (((int)(val / 26)) - 1);
            int i2 = ((val % 26) - 1);
            char ch1 = Translate0To25(i1);
            char ch2 = Translate0To25(i2);
            return new string(new char[]{ch1, ch2});
        }
    }

    internal static string AlignCellString(CCell cell, CCellLayout layout){
        Check.Reference(cell);
        Check.Reference(layout);
        Check.Expression(layout.IsFixedWidth);
        string s = cell.ToString();
        int colWidth = layout.FixedWidth;
        CellAlignment align = layout.Alignment;
        if(CellAlignment.Auto == align){
            switch(cell.ValueFamily){
                case CellValueFamily.Numeric:
                    align = CellAlignment.Right;
                    break;
                default:
                    align = CellAlignment.Left;
                    break;
            }
        }
        switch(align){
            case CellAlignment.Left:
                if(s.Length > colWidth){
                    s = ApplyOverflow(s, layout.Overflow, 0, colWidth);
                }else if(s.Length < colWidth){
                    s = s.PadRight(colWidth);
                }
                break;
            case CellAlignment.Center:
                if(s.Length > colWidth){
                    int n1, n2;
                    SplitInt(s.Length - colWidth, out n1, out n2);
                    s = ApplyOverflow(s, layout.Overflow, n1, colWidth);
                }else if(s.Length < colWidth){
                    int n1, n2;
                    SplitInt(colWidth - s.Length, out n1, out n2);
                    s = s.PadLeft(s.Length + n1);
                    s = s.PadRight(s.Length + n2);
                }
                break;
            case CellAlignment.Right:
                if(s.Length > colWidth){
                    s = ApplyOverflow(s, layout.Overflow, s.Length - colWidth, colWidth);
                }else if(s.Length < colWidth){
                    s = s.PadLeft(colWidth);
                }
                break;
        }
        return s;
    }

    private static char Translate0To25(int i){
        int ascA = (int) 'A';
        int asc = (ascA + i);
        return (char) asc;
    }

    private static void SplitInt(int n, out int n1, out int n2){
        n1 = (int) n / 2;
        n2 = n - n1;
    }

    private static string ApplyOverflow(string s, CellOverflow of, int iStart, int nMaxLength){
        switch(of){
            case CellOverflow.Blackout:
                s = new string('#', nMaxLength);
                break;
            case CellOverflow.Limit:
                if(nMaxLength >= 4){
                    s = Parse.LimitText(s, nMaxLength - 4, " ...");                    
                }else{
                    s = new string('.', nMaxLength);
                }
                break;
            case CellOverflow.Truncate:
                s = s.Substring(iStart, nMaxLength);
                break;
        }
        return s;
    }

} // class CellUtil

public class CSheetException : Exception {

    public CSheetException(string sMessage) : base(sMessage) {
        return;
    }

    public override string ToString(){
        return  "#" + this.Message;
    }

} // class CSheetException

public interface IRange {

    CCell this[int iRow, int iCol]{get;}
    CCell this[string sCoords]{get;}
    int RowCount{get;}
    int ColumnCount{get;}

} // interface IRange

public class CSheet : IRange {

    private const int EXTRA_GROW = 0;

    private CCell[,] m_Cells;
    private CFunctionPool m_FunctionPool;
    private CSheetInterpreter m_Interpreter;
    private Stack m_EvalStack;

    public CSheet(){
        m_Cells = new CCell[0, 0];
        m_FunctionPool = new CFunctionPool(this);
        m_Interpreter = new CSheetInterpreter(this);
        m_EvalStack = new Stack();
    }

    public CCell this[int iRow, int iCol]{
        get{
            int iMaxRow = m_Cells.GetUpperBound(0);
            int iMaxCol = m_Cells.GetUpperBound(1);
            if(iRow > iMaxRow || iCol > iMaxCol){
                int cRows = (iRow > iMaxRow ? iRow : iMaxRow) + 1 + EXTRA_GROW;
                int cCols = (iCol > iMaxCol ? iCol : iMaxCol) + 1 + EXTRA_GROW;
                CCell[,] newCells = new CCell[cRows, cCols];
                for(int i = 0; i <= iMaxRow; i++){
                    for(int j = 0; j <= iMaxCol; j++){
                        newCells[i, j] = m_Cells[i, j];
                    }
                }
                m_Cells = newCells;
            }
            if(m_Cells[iRow, iCol] == null){
                m_Cells[iRow, iCol] = new CCell(this, iRow, iCol);
            }
            return m_Cells[iRow, iCol];
        }
    }

    public CCell this[string sCoords]{
        get{
            int iRow = -1, iCol = -1;
            CellUtil.ParseCoordinates(sCoords, ref iRow, ref iCol);
            return this[iRow, iCol];
        }
    }

    public CFunctionPool FunctionPool{
        get{
            return m_FunctionPool;
        }
    }

    public CInterpreter Interpreter{
        get{
            return m_Interpreter;
        }
    }

    public int RowCount{
        get{
            return m_Cells.GetUpperBound(0) + 1;
        }
    }

    public int ColumnCount{
        get{
            return m_Cells.GetUpperBound(1) + 1;
        }
    }

    public string ToHtml(){
        StringBuilder sb = new StringBuilder();
        sb.Append(Environment.NewLine);
        sb.Append("<table width=\"100%\" border=\"1\">" + Environment.NewLine);
        sb.Append("<thead>");
        sb.Append("<th bgcolor=\"silver\">&nbsp;</th>");
        int iMaxCol = m_Cells.GetUpperBound(1);
        for(int iCol = 0; iCol <= iMaxCol; iCol++){
            sb.Append("<th bgcolor=\"silver\">" + CellUtil.TranslateColumn(iCol) + "</th>");
        }
        sb.Append("</thead>" + Environment.NewLine);
        sb.Append(this.ToString("<tr>",
                                true, "<td bgcolor=\"silver\">", 1, "</td>",
                                "<td>", null, "</td>",
                                "</tr>" + Environment.NewLine));
        sb.Append("</table>" + Environment.NewLine);
        return sb.ToString();
    }

    public override string ToString(){
        StringBuilder sb = new StringBuilder();
        int iMaxRow = m_Cells.GetUpperBound(0);
        int iMaxCol = m_Cells.GetUpperBound(1);
        sb.Append("CSheet rows: " + (iMaxRow + 1) + ", columns: " + (iMaxCol + 1) + Environment.NewLine);
        sb.Append("-------------------------------------------------------------------------------" + Environment.NewLine);
        CCellLayout layout = new CCellLayout(5, 15);
        sb.Append(this.ToString(string.Empty,
                                false, string.Empty, 0, string.Empty,
                                string.Empty, layout, " ",
                                Environment.NewLine));
        sb.Append("-------------------------------------------------------------------------------" + Environment.NewLine);
        return sb.ToString();
    }

    public string ToString(string sLineBegin,
                           bool fUseRowNo, string sRowPrefixBegin, int nRowOffset, string sRowPrefixEnd,
                           string sCellBegin, CCellLayout layout, string sCellEnd,
                           string sLineEnd){
        StringBuilder sb = new StringBuilder();
        int iMaxRow = m_Cells.GetUpperBound(0);
        int iMaxCol = m_Cells.GetUpperBound(1);
        CCellLayout[] colLayouts = null;
        if(layout != null){
            if(layout.IsFixedWidth){
                colLayouts = new CCellLayout[iMaxCol + 1];
                for(int iCol = 0; iCol <= iMaxCol; iCol++){
                    colLayouts[iCol] = layout;
                }
            }else{
                colLayouts = new CCellLayout[iMaxCol + 1];
                for(int iCol = 0; iCol <= iMaxCol; iCol++){
                    int colWidth = 0;
                    for(int iRow = 0; iRow <= iMaxRow; iRow++){
                        string s = this.GetCellString(iRow, iCol);
                        if(s.Length > colWidth){
                            colWidth = s.Length;
                        }
                    }
                    if(colWidth < layout.MinWidth){
                        colWidth = layout.MinWidth;
                    }else if(colWidth > layout.MaxWidth){
                        colWidth = layout.MaxWidth;
                    }
                    colLayouts[iCol] = new CCellLayout(colWidth);
                    colLayouts[iCol].Alignment = layout.Alignment;
                    colLayouts[iCol].Overflow = layout.Overflow;
                }
            }
        }
        for(int iRow = 0; iRow <= iMaxRow; iRow++){
            sb.Append(sLineBegin);
            if(fUseRowNo){
                sb.Append(sRowPrefixBegin);
                sb.Append((iRow + nRowOffset));
                sb.Append(sRowPrefixEnd);
            }
            for(int iCol = 0; iCol <= iMaxCol; iCol++){
                sb.Append(sCellBegin);
                if(colLayouts != null){
                    sb.Append(CellUtil.AlignCellString(this[iRow, iCol], colLayouts[iCol]));
                }else{
                    sb.Append(this.GetCellString(iRow, iCol));
                }
                sb.Append(sCellEnd);
            }
            sb.Append(sLineEnd);
        }
        return sb.ToString();
    }

    public XmlElement ToXml(XmlDocument xd){
        XmlElement xe = xd.CreateElement("sheet");
        foreach(CCell cell in m_Cells){
            if(cell != null){
                XmlElement xeChild = cell.ToXml(xd);
                xe.AppendChild(xeChild);
            }
        }
        string sCode = this.Interpreter.GetPersistableCode();
        sCode = Consts.CRLF_DBL + sCode;
        XmlCDataSection cdata = xd.CreateCDataSection(sCode);
        xe.AppendChild(xd.CreateTextNode(Environment.NewLine));
        xe.AppendChild(cdata);
        xe.AppendChild(xd.CreateTextNode(Environment.NewLine));
        return xe;
    }

    public void FromXml(XmlElement xe){
        this.Clear();
        foreach(XmlNode xeChild in xe){
            switch(xeChild.NodeType){
                case XmlNodeType.Element:
                    string sRow = xeChild.Attributes["row"].Value;
                    string sCol = xeChild.Attributes["col"].Value;
                    string sText = xeChild.Attributes["text"].Value;
                    int iRow = int.Parse(sRow);
                    int iCol = int.Parse(sCol);
                    this[iRow, iCol].Text = sText;
                    break;
                case XmlNodeType.CDATA:
                    try{
                        string sCode = xeChild.InnerText;
                        this.Interpreter.Evaluate(sCode);
                    }catch(Exception ex){
                        Console.WriteLine(ex.Message);
                    }
                    break;
            }
        }
    }

    public void Clear(){
        m_Cells = new CCell[0, 0];
        m_Interpreter.ClearObjects();
    }

    internal object Evaluate(CCell cell){
        object ret = null;
        string sText = cell.Text;
        if(sText != null){
            // avoid self or circular references
            object[] aoCells = m_EvalStack.ToArray();
            foreach(CCell c in aoCells){
                if(c.Equals(cell)){
                    if(aoCells.Length == 1){
                        throw new CSheetException("Self-reference.");
                    }else{
                        throw new CSheetException("Circular reference.");
                    }
                }
            }
            // so
            try{
                m_EvalStack.Push(cell);
                IResult res = m_Interpreter.Evaluate(cell.Text);
                ret = res.Value; // null ok for void
            }catch(CSheetException ex){
                ret = ex;
            }catch(CInterpreterException ex){
                Exception exInnerMost = this.GetInnerMostException(ex);
                ret = new CSheetException(exInnerMost.Message);
            }catch(Exception ex){
                Exception exInnerMost = this.GetInnerMostException(ex);
                ret = new CSheetException(exInnerMost.Message);
            }finally{
                m_EvalStack.Pop();
            }
        }
        return ret;
    }

    private Exception GetInnerMostException(Exception ex){
        while(ex.InnerException != null){
            ex = ex.InnerException; 
        }
        return ex;
    }

    private bool IsCell(int iRow, int iCol){
        bool fRet = false;
        if(iRow < this.RowCount && iCol < this.ColumnCount){
            fRet = (m_Cells[iRow, iCol] != null);
        }
        return fRet;
    }

    private string GetCellString(int iRow, int iCol){
        string sRet = string.Empty;
        if(iRow < this.RowCount && iCol < this.ColumnCount){
            CCell cell = m_Cells[iRow, iCol];
            if(cell != null){
                sRet = cell.ToString();
            }
        }
        return sRet;
    }

} // class CSheet

public class CCell {

    private CSheet m_Sheet;
    private int m_RowIndex;
    private int m_ColIndex;
    private string m_Text;

    internal CCell(CSheet sheet, int iRow, int iCol){
        m_Sheet = sheet;
        m_RowIndex = iRow;
        m_ColIndex = iCol;
        m_Text = null;
    }

    public CSheet Sheet{
        get{
            return m_Sheet;
        }
    }

    public int RowIndex{
        get{
            return m_RowIndex;
        }
    }

    public int ColumnIndex{
        get{
            return m_ColIndex;
        }
    }

    public string Text{
        get{
            return m_Text;
        }
        set{
            m_Text = value;
        }
    }

    public object Value{
        get{
            return m_Sheet.Evaluate(this);
        }
    }

    public CellValueFamily ValueFamily{
        get{
            object val = this.Value;
            if(null == val){
                return CellValueFamily.Null;
            }else if(val is string){
                return CellValueFamily.Text;
            }else if(CellUtil.IsNumber(val)){
                return CellValueFamily.Numeric;
            }else if(val is Exception){
                return CellValueFamily.Error;
            }else{
                return CellValueFamily.Object;
            }
        }
    }

    public override string ToString(){
        object value = this.Value;
        if(value != null){
            if(value is double){
                return value.ToString(); // ...
            }else{
                return value.ToString();
            }
        }else{
            return string.Empty; // "(null)";
        }
        // return "{" + m_RowIndex + "." + m_ColIndex + " " + this.Text + "}";
    }

    public override bool Equals(object obj){
        if(obj != null && obj.GetType().Equals(this.GetType())){
            CCell that = (CCell) obj;
            return this.RowIndex == that.RowIndex && this.ColumnIndex == that.ColumnIndex;
        }else{
            return false;
        }
    }

    public override int GetHashCode(){
        return m_RowIndex.GetHashCode();
    }

    public XmlElement ToXml(XmlDocument xd){
        XmlElement xe = xd.CreateElement("cell");
        XmlAttribute xa = xd.CreateAttribute("row");
        xa.Value = m_RowIndex.ToString();
        xe.Attributes.Append(xa);
        xa = xd.CreateAttribute("col");
        xa.Value = m_ColIndex.ToString();
        xe.Attributes.Append(xa);
        xa = xd.CreateAttribute("text");
        xa.Value = (m_Text != null ? m_Text : string.Empty);
        xe.Attributes.Append(xa);
        return xe;
    }

} // class CCell

public enum CellValueFamily {
    Null,
    Numeric,
    Text,
    Object,
    Error
}

public enum CellAlignment {
    Auto,
    Left,
    Center,
    Right
}

public enum CellOverflow {
    Blackout,
    Limit,
    Truncate
}

public class CCellLayout {

    private int m_FixedWidth;
    private int m_MinWidth;
    private int m_MaxWidth;
    private CellAlignment m_Alignment;
    private CellOverflow m_Overflow;

    private CCellLayout(){
        m_Alignment = CellAlignment.Auto;
        m_Overflow = CellOverflow.Limit;
    }

    public CCellLayout(int fixedWidth) : this() {
        Check.ValueGreaterOrEqual(fixedWidth, 0);
        m_FixedWidth = fixedWidth;
        m_MinWidth = fixedWidth;
        m_MaxWidth = fixedWidth;
    }

    public CCellLayout(int minWidth, int maxWidth) : this() {
        Check.ValueGreaterOrEqual(minWidth, 0);
        Check.ValueGreaterOrEqual(maxWidth, minWidth);
        m_FixedWidth = -1;
        m_MinWidth = minWidth;
        m_MaxWidth = maxWidth;
    }

    public int FixedWidth{
        get{
            return m_FixedWidth;
        }
    }

    public int MinWidth{
        get{
            return m_MinWidth;
        }
    }

    public int MaxWidth{
        get{
            return m_MaxWidth;
        }
    }

    public CellAlignment Alignment{
        get{
            return m_Alignment;
        }
        set{
            m_Alignment = value;
        }
    }

    public CellOverflow Overflow{
        get{
            return m_Overflow;
        }
        set{
            m_Overflow = value;
        }
    }

    public bool IsFixedWidth{
        get{
            return m_FixedWidth >= 0;
        }
    }

} // class CCellLayout

} // namespace Gregor.NetSheet
