using System;
using System.Collections;
using System.Reflection;
using System.Text;
using Gregor.Core;
using Gregor.NetConsole;
using Gregor.NetConsole.Engine;
using Gregor.NetConsole.Parser;

namespace Gregor.NetSheet {

public class CSheetInterpreter : CCurlyInterpreter {

    private CSheet m_Sheet;

    public CSheetInterpreter(CSheet sheet) : base() {
        Check.Reference(sheet);
        m_Sheet = sheet;
        this.OptionNoCase = true;
        this.ImportNamespace("System");
        this.ImportNamespace("Gregor.Core");
        this.ImportNamespace("Gregor.NetSheet");
    }

    public CSheet Sheet{
        get{
            return m_Sheet;
        }
    }

    protected override void CheckVariableNameAndAccess(string s, bool fForWriting, bool fExplicitDeclaration){
        if(CellUtil.AreValidCoordinates(s)){
            throw new CInterpreterException("Cannot use a cell reference as a variable or function name: '" + s + "' (use the 'set' command for editing cells).");
        }else if(Flow.IsAmong(s, this.OptionNoCase, "loadxml", "savexml", "loadcode", "get", "set", "html", "new", "eval", "help")){
            throw new CInterpreterException("Bad variable name.");
        }
        base.CheckVariableNameAndAccess(s, fForWriting, fExplicitDeclaration);
    }

    protected override IResult EvaluateUnqualifiedCallExpression(CEvaluationContext context, CMethodCallExpression methodCallExpr){
        IResult ret = CResult.Void;
        if(methodCallExpr.Children[0].NodeType != CodeNodeType.MemberReference){
            throw new CInterpreterException("Name of function expected.", methodCallExpr);
        }
        CMemberReference memberRef = (CMemberReference) methodCallExpr.Children[0];
        string sMemberName = memberRef.GetMemberName();
        NResults arguments = this.GetArgumentsOfCallExpression(context, methodCallExpr, null);
        // cell-access shortcut?
        if(CellUtil.AreValidCoordinates(sMemberName)){
            if(arguments.Count > 0){
                throw new CInterpreterException("No arguments expected in cell-access.", methodCallExpr);
            }
            CCell cell = this.Sheet[sMemberName];
            ret = (cell.Value != null ? new CResult(cell.Value, cell.Value.GetType()) : CResult.Null);
        }else{
            // .NET Sheet function?
            CFunctionPool pool = m_Sheet.FunctionPool;
            Type tp = pool.GetType();
            IResult me = new CResult(pool, tp);
            MemberInfo mbi = this.TryGetMember(tp, sMemberName, arguments, false);
            if(mbi != null){
                ret = this.GetValue(me, mbi, arguments);
            }else{
                // assume a real user function
                ret = base.EvaluateUnqualifiedCallExpression(context, methodCallExpr);
            }
        }
        // so
        return ret;
    }

    private MemberInfo TryGetMember(Type tp, string sMemberName, NResults arguments, bool fStatic){
        try{
            // (Type tp, string sMemberName, MemberTypes memberTypes, Type[] genericTypeArguments, NResults arguments, MemberScope scope)
            return this.GetMember(tp, sMemberName, MemberTypes.Method, Type.EmptyTypes, arguments, MemberScope.Instance);
        }catch(Exception){
            return null;
        }
    }

} // class CEvaluationContext

} // namespace Gregor.NetSheet
