using System;
using System.Collections;
using System.Reflection;
using System.Text;
using Gregor.Core;

namespace Gregor.NetSheet {

internal delegate object RangeCallback(object result, CCell cell);

public class CFunctionPool {


    #region "Nested types"
    private class CAvgData{

        private double m_Sum;
        private int m_Count;

        public CAvgData(double initialValue){
            m_Sum = initialValue;
            m_Count = 0;
        }

        public double Result{
            get{
                return m_Sum / m_Count;
            }
        }

        public void Add(double v){
            m_Sum += v;
            m_Count++;
        }
    }
    #endregion


    #region "Instance Fields"
    private CSheet m_Sheet;
    #endregion


    #region "Constructors"
    public CFunctionPool(CSheet sheet){
        Check.Reference(sheet);
        m_Sheet = sheet;
    }
    #endregion


    #region "Single-cell access functions"
    public object Cell(int iRow, int iCol){
        CCell cell = m_Sheet[iRow, iCol];
        return cell.Value;
    }
    public object Cell(string sCoords){
        CCell cell = m_Sheet[sCoords];
        return cell.Value;
    }
    public string GetCellExpression(int iRow, int iCol){
        CCell cell = m_Sheet[iRow, iCol];
        return cell.Text;
    }
    public string GetCellExpression(string sCoords){
        CCell cell = m_Sheet[sCoords];
        return cell.Text;
    }
    public void SetCellExpression(int iRow, int iCol, string sExpr){
        CCell cell = m_Sheet[iRow, iCol];
        cell.Text = sExpr;
    }
    public void SetCellExpression(string sCoords, string sExpr){
        CCell cell = m_Sheet[sCoords];
        cell.Text = sExpr;
    }
    #endregion


    #region "Range functions"
    public object Sum(string sCoordsFrom, string sCoordsTo){
        try{
            return ProcessRange(sCoordsFrom, sCoordsTo, new RangeCallback(SumItUp));
        }catch(ArgumentException ex){
            return ex.Message;
        }
    }

    private object SumItUp(object result, CCell cell){
        double val = 0.0;
        if(result != null){
            val = (double) result;
        }
        object o = cell.Value;
        if(o != null){
            double v = 0.0;
            if(false == CellUtil.TryGetNumber(o, ref v)){
                throw new CSheetException("NaN: " + o);
            }
            val += v;
        }
        return val;
    }

    public object Avg(string sCoordsFrom, string sCoordsTo){
        try{
            CAvgData data = (CAvgData) ProcessRange(sCoordsFrom, sCoordsTo, new RangeCallback(AvgItUp));
            return data.Result;
        }catch(ArgumentException ex){
            return ex.Message;
        }
    }

    private object AvgItUp(object result, CCell cell){
        CAvgData data = (CAvgData) result;
        if(result == null){
            data = new CAvgData(0.0);
        }
        object o = cell.Value;
        if(o != null){
            double v = 0.0;
            if(false == CellUtil.TryGetNumber(o, ref v)){
                throw new CSheetException("NaN: " + o);
            }
            data.Add(v);
        }
        return data;
    }

    private object ProcessRange(string sCoordsFrom, string sCoordsTo, RangeCallback callback){
        object result = null;
        CCell cellFrom = m_Sheet[sCoordsFrom];
        CCell cellTo = m_Sheet[sCoordsTo];
        if(cellFrom.RowIndex > cellTo.RowIndex){
            throw new CSheetException("Bad range.");
        }
        if(cellFrom.ColumnIndex > cellTo.ColumnIndex){
            throw new CSheetException("Bad range.");
        }
        for(int iCol = cellFrom.ColumnIndex; iCol <= cellTo.ColumnIndex; iCol++){
            for(int iRow = cellFrom.RowIndex; iRow <= cellTo.RowIndex; iRow++){
                CCell cell = m_Sheet[iRow, iCol];
                result = callback(result, cell);
            }
        }
        return result;
    }
    #endregion


    #region "Variable arguments functions"
    public object Add(params object[] values){
        double val = 0.0;
        foreach(object o in values){
            if(o != null){
                double v = 0.0;
                if(false == CellUtil.TryGetNumber(o, ref v)){
                    throw new CSheetException("NaN: " + o);
                }
                val += v;
            }
        }
        return val;
    }

    public object Multiply(params object[] values){
        double val = 1.0;
        foreach(object o in values){
            if(o != null){
                double v = 0.0;
                if(false == CellUtil.TryGetNumber(o, ref v)){
                    throw new CSheetException("NaN: " + o);
                }
                val *= v;
            }
        }
        return val;
    }

    public object Concat(params object[] values){
        StringBuilder sb = new StringBuilder();
        foreach(object o in values){
            if(o != null){
                sb.Append(o.ToString());
            }
        }
        return sb.ToString();
    }
    #endregion


    #region "Other functions"
    public object Pow(object baese, object power){
        double dBase = 0.0, dPower = 0.0;
        if(false == CellUtil.TryGetNumber(baese, ref dBase)){
            throw new CSheetException("NaN: " + baese);
        }
        if(false == CellUtil.TryGetNumber(power, ref dPower)){
            throw new CSheetException("NaN: " + power);
        }
        return Math.Pow(dBase, dPower);
    }
    #endregion


} // class CFunctionPool

} // namespace Gregor.NetSheet
