// <clown-build> csc.exe /debug /d:DEBUG /out:NetSheet.exe /r:Gregor.Core.dll,Gregor.NetConsole.dll NetSheet.*.cs </clown-build>

using System;
using System.IO;
using System.Text;
using System.Xml;
using Gregor.Core;
using Gregor.NetConsole;

namespace Gregor.NetSheet {

[Module()]
public class Start {

    public static void Main(string[] args){

        Dev.TempFileNameBase = "NetSheet";
        Dev.TempFileExtension = ".html";

        if(args.Length == 1 && args[0] == "/help"){
            SyntaxExit();
        }

        try{
            Run(args);
        }catch(Exception ex){
            Dev.ProcessException(ex);
            Environment.Exit(1);
        }
    }

    private static void Run(string[] args){

        Console.Title = ".NET Sheet";

        CNetSheet netSheet = new CNetSheet();
        netSheet.Input += new InputEventHandler(HandleInput);
        netSheet.Log += new LogEventHandler(HandleLog);

        if(false == netSheet.TryEvalInput(args)){
            Environment.Exit(1);
        }
        netSheet.RunLoop();
    }

    private static void SyntaxExit(){
        Dev.Trace(".NET Sheet - Interactive Spreadsheet");
        Dev.Trace("Syntax:");
        Dev.Trace("    NetSheet [/help] ^ [<command>+]");
        Environment.Exit(2);
    }

    private static void HandleInput(object sender, CInputEventArgs e){
        try{
            Console.Write(".NET Sheet>");
            e.Input = Console.ReadLine().Trim();
        }catch(Exception ex){
            Dev.ProcessException(ex);
        }
    }

    private static void HandleLog(object sender, CLogEventArgs e){
        try{
            Console.WriteLine(e.ToString());
        }catch(Exception ex){
            Dev.ProcessException(ex, true);
        }
    }

} // module Start

public class CNetSheet : CNetConsole {

    private CSheet m_Sheet;

    public CNetSheet() : base() {
        m_Sheet = new CSheet();
        this.SetInterpreter(m_Sheet.Interpreter);
    }

    public CSheet Sheet{
        get{
            return m_Sheet;
        }
    }

    public CSheetInterpreter SheetInterpreter{
        get{
            return (CSheetInterpreter) this.Interpreter;
        }
    }

    public bool TryEvalInput(string[] args){
        foreach(string s in args){
            if(false == this.TryEvalInput(s)){
                return false;
            }
        }
        return true;
    }

    public override bool TryEvalInput(string s){
        try{
            if(false == Flow.IsString(s)){
                return true;
            }
            else if(0 == string.Compare(s, "exit", m_Sheet.Interpreter.OptionNoCase)){
                return false;
            }
            else if(0 == string.Compare(s, "help", m_Sheet.Interpreter.OptionNoCase)){
                this.ShowHelp();
                return true;
            }
            else if(0 == string.Compare(s, "html", m_Sheet.Interpreter.OptionNoCase)){
                this.ShellHtml();
                return true;
            }
            else if(Parse.StringStartsWith(s, "set ", m_Sheet.Interpreter.OptionNoCase)){
                string sErrMsg = "Assignment to cell expected.";
                int iEq = s.IndexOf('=', 4);
                Check.ValueGreaterOrEqual(iEq, 4, sErrMsg);
                string sCoords = s.Substring(4, iEq - 4).Trim();
                Check.Expression(CellUtil.AreValidCoordinates(sCoords), sErrMsg);
                string sValue = s.Substring(iEq + 1).Trim();
                Check.String(sValue, sErrMsg);
                m_Sheet[sCoords].Text = sValue;
                return true;
            }
            else if(Parse.StringStartsWith(s, "get ", m_Sheet.Interpreter.OptionNoCase)){
                string sErrMsg = "Cell reference expected.";
                string sCoords = s.Substring(4).Trim();
                Check.Expression(CellUtil.AreValidCoordinates(sCoords), sErrMsg);
                W(m_Sheet[sCoords].Text);
                return true;
            }
            else if(Parse.StringStartsWith(s, "eval ", m_Sheet.Interpreter.OptionNoCase)){
                string sErrMsg = "Cell reference expected.";
                string sCoords = s.Substring(5).Trim();
                Check.Expression(CellUtil.AreValidCoordinates(sCoords), sErrMsg);
                W(m_Sheet[sCoords].Value);
                return true;
            }
            else if(0 == string.Compare(s, "eval", m_Sheet.Interpreter.OptionNoCase)){
                Console.Write(m_Sheet);
                return true;
            }
            else if(0 == string.Compare(s, "new", m_Sheet.Interpreter.OptionNoCase)){
                m_Sheet.Clear();
                return true;
            }
            else if(Parse.StringStartsWith(s, "loadxml ", m_Sheet.Interpreter.OptionNoCase)){
                string sFile = s.Substring(8).Trim();
                Check.String(sFile, "File name expected.");
                XmlDocument xd = Dev.ReadXmlFile(sFile);
                m_Sheet.FromXml(xd.DocumentElement);
                return true;
            }
            else if(Parse.StringStartsWith(s, "savexml ", m_Sheet.Interpreter.OptionNoCase)){
                string sFile = s.Substring(8).Trim();
                Check.String(sFile, "File name expected.");
                XmlDocument xd = new XmlDocument();
                xd.AppendChild(m_Sheet.ToXml(xd));
                Dev.WriteXmlFile(sFile, xd);
                return true;
            }
            else if(Parse.StringStartsWith(s, "loadcode ", m_Sheet.Interpreter.OptionNoCase)){
                string sFile = s.Substring(9).Trim();
                Check.String(sFile, "File name expected.");
                this.LoadCode(sFile);
                return true;
            }
            else{
                return base.TryEvalInput(s);
            }
        }catch(Exception ex){
            W(ex.Message);
            return true;
        }
    }

    public override void ShowHelp(){
        W(string.Empty);
        W(".NET Sheet - Interactive Spreadsheet");
        W(string.Empty);
        W("Commands:");
        W(string.Empty);
        W("    General:");
        W(string.Empty);
        W("        help");
        W("        exit");
        W("        new");
        W("        loadxml <xml-file>");
        W("        savexml <xml-file>");
        W("        loadcode <cs-file>");
        W("        html // shell HTML file");
        W("        most .NET Console commands");
        W(string.Empty);
        W("    Cell Editing and Formula Viewing:");
        W(string.Empty);
        W("        set <cell-ref> = <expression>");
        W("        get <cell-ref>");
        W(string.Empty);
        W("    Evaluation:");
        W(string.Empty);
        W("        eval <cell-ref>");
        W("        eval // whole sheet");
        W(string.Empty);
        W("Expressions:");
        W(string.Empty);
        W("    Literals:");
        W(string.Empty);
        W("        booleans    // true | false");
        W("        chars       // 'a'");
        W("        strings     // \"Hello\"");
        W("        ints        // 42");
        W("        doubles     // 66.6");
        W("        nulls       // null");
        W(string.Empty);
        W("    Cell Evaluation:");
        W(string.Empty);
        W("        Cell(<int>, <int>) // zero-based row, col like Cell(0, 0)");
        W("        Cell(<string>)     // alphanumerics <col><row> like Cell(\"A1\")");
        W(string.Empty);
        W("    Cell References:");
        W(string.Empty);
        W("        // to pass cells to range functions by reference (without");
        W("        // evaluation), use strings, ig. Sum(\"A1\", \"A5\")");
        W(string.Empty);
        W("    Range Functions:");
        W(string.Empty);
        W("        Sum(<cell-ref-string>, <cell-ref-string>");
        W("        Avg(<cell-ref-string>, <cell-ref-string>");
        W(string.Empty);
        W("    Variable Arguments:");
        W(string.Empty);
        W("        Add(<expressions> ...)");
        W("        Multiply(<expressions> ...)");
        W("        Concat(<expressions> ...)");
        W(string.Empty);
        W("    Other:");
        W(string.Empty);
        W("        Pow(number, number)");
        W("        <any .NET method>");
        W(string.Empty);
    }

    private void W(object o){
        this.OnLog(new CLogEventArgs(Conv.ObjectToString(o, null)));
    }

    public void ShellHtml(){
        StringBuilder sb = new StringBuilder();
        sb.Append("<html>" + Environment.NewLine + "<head><title>.NET Sheet</title></head>" + Environment.NewLine + "<body>" + Environment.NewLine);
        sb.Append(m_Sheet.ToHtml());
        sb.Append(Environment.NewLine + "</body>" + Environment.NewLine + "</html>" + Environment.NewLine);
        Dev.ShellTemporaryFile(sb.ToString());
    }

} // class CNetSheet

} // namespace Gregor.NetSheet
